(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('immer-peasy'), require('react'), require('redux'), require('redux-thunk'), require('memoizerific')) :
  typeof define === 'function' && define.amd ? define(['exports', 'immer-peasy', 'react', 'redux', 'redux-thunk', 'memoizerific'], factory) :
  (global = global || self, factory(global.EasyPeasy = {}, global.produce, global.React, global.Redux, global.ReduxThunk, global.memoizerific));
}(this, function (exports, immerPeasy, React, redux, reduxThunk, memoizerific) { 'use strict';

  var React__default = 'default' in React ? React['default'] : React;
  reduxThunk = reduxThunk && reduxThunk.hasOwnProperty('default') ? reduxThunk['default'] : reduxThunk;
  memoizerific = memoizerific && memoizerific.hasOwnProperty('default') ? memoizerific['default'] : memoizerific;

  var StoreContext = React.createContext();

  // To get around it, we can conditionally useEffect on the server (no-op) and
  // useLayoutEffect in the browser. We need useLayoutEffect to ensure the store
  // subscription callback always has the selector from the latest render commit
  // available, otherwise a store update may happen between render and the effect,
  // which may cause missed updates; we also must ensure the store subscription
  // is created synchronously, otherwise a store update may occur before the
  // subscription is created and an inconsistent state may be observed

  var useIsomorphicLayoutEffect = typeof window !== 'undefined' ? React.useLayoutEffect : React.useEffect;
  function createStoreStateHook(Context) {
    return function useStoreState(mapState) {
      var store = React.useContext(Context);
      var mapStateRef = React.useRef(mapState);
      var stateRef = React.useRef();
      var subscriptionMapStateError = React.useRef();

      var _useReducer = React.useReducer(function (s) {
        return s + 1;
      }, 0),
          forceRender = _useReducer[1];

      if (subscriptionMapStateError.current || mapStateRef.current !== mapState || stateRef.current === undefined) {
        try {
          stateRef.current = mapState(store.getState());
        } catch (err) {
          var errorMessage = "An error occurred trying to map state in a useStoreState hook: " + err.message + ".";

          if (subscriptionMapStateError.current) {
            errorMessage += "\nThis error may be related to the following error:\n" + subscriptionMapStateError.current.stack + "\n\nOriginal stack trace:";
          }

          throw new Error(errorMessage);
        }
      }

      useIsomorphicLayoutEffect(function () {
        mapStateRef.current = mapState;
        subscriptionMapStateError.current = undefined;
      });
      useIsomorphicLayoutEffect(function () {
        var checkMapState = function checkMapState() {
          try {
            var newState = mapStateRef.current(store.getState());

            if (newState === stateRef.current) {
              return;
            }

            stateRef.current = newState;
          } catch (err) {
            // see https://github.com/reduxjs/react-redux/issues/1179
            // There is a possibility mapState will fail due to stale state or
            // props, therefore we will just track the error and force our
            // component to update. It should then receive the updated state
            subscriptionMapStateError.current = err;
          }

          forceRender({});
        };

        var unsubscribe = store.subscribe(checkMapState);
        checkMapState();
        return unsubscribe;
      }, []);
      return stateRef.current;
    };
  }
  var useStoreState = createStoreStateHook(StoreContext);
  function createStoreActionsHook(Context) {
    return function useStoreActions(mapActions) {
      var store = React.useContext(Context);
      return mapActions(store.getActions());
    };
  }
  var useStoreActions = createStoreActionsHook(StoreContext);
  function createStoreDispatchHook(Context) {
    return function useStoreDispatch() {
      var store = React.useContext(Context);
      return store.dispatch;
    };
  }
  var useStoreDispatch = createStoreDispatchHook(StoreContext);
  function createTypedHooks() {
    return {
      useStoreActions: useStoreActions,
      useStoreDispatch: useStoreDispatch,
      useStoreState: useStoreState
    };
  }

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  var actionSymbol = '🙈action🙈';
  var actionOnSymbol = '🙈actionOn🙈';
  var computedSymbol = '🙈computedSymbol🙈';
  var reducerSymbol = '🙈reducer🙈';
  var thunkOnSymbol = '🙈thunkOn🙈';
  var thunkSymbol = '🙈thunk🙈';

  var debug = function debug(state) {
    if (immerPeasy.isDraft(state)) {
      return immerPeasy.finishDraft(immerPeasy.createDraft(state));
    }

    return state;
  };
  var memo = function memo(fn, cacheSize) {
    return memoizerific(cacheSize)(fn);
  };
  var actionOn = function actionOn(targetResolver, fn) {
    fn[actionOnSymbol] = {
      targetResolver: targetResolver
    };
    return fn;
  };
  var action = function action(fn) {
    fn[actionSymbol] = {};
    return fn;
  };
  var defaultStateResolvers = [function (state) {
    return state;
  }];
  var computed = function computed(fnOrStateResolvers, fn) {
    if (typeof fn === 'function') {
      fn[computedSymbol] = {
        stateResolvers: fnOrStateResolvers
      };
      return fn;
    }

    fnOrStateResolvers[computedSymbol] = {
      stateResolvers: defaultStateResolvers
    };
    return fnOrStateResolvers;
  };
  var thunkOn = function thunkOn(targetResolver, fn) {
    fn[thunkOnSymbol] = {
      targetResolver: targetResolver
    };
    return fn;
  };
  var thunk = function thunk(fn) {
    fn[thunkSymbol] = {};
    return fn;
  };
  var reducer = function reducer(fn) {
    fn[reducerSymbol] = {};
    return fn;
  };

  var isStateObject = function isStateObject(x) {
    return x !== null && typeof x === 'object' && !Array.isArray(x) && x.constructor === Object;
  };
  var get = function get(path, target) {
    return path.reduce(function (acc, cur) {
      return isStateObject(acc) ? acc[cur] : undefined;
    }, target);
  };
  var set = function set(path, target, value) {
    path.reduce(function (acc, cur, idx) {
      if (idx + 1 === path.length) {
        acc[cur] = value;
      } else {
        acc[cur] = acc[cur] || {};
      }

      return acc[cur];
    }, target);
  };

  var newify = function newify(currentPath, currentState, finalValue) {
    if (currentPath.length === 0) {
      return finalValue;
    }

    var newState = _extends({}, currentState);

    var key = currentPath[0];

    if (currentPath.length === 1) {
      newState[key] = finalValue;
    } else {
      newState[key] = newify(currentPath.slice(1), newState[key], finalValue);
    }

    return newState;
  };

  function createStoreInternals(_ref) {
    var disableImmer = _ref.disableImmer,
        initialState = _ref.initialState,
        injections = _ref.injections,
        model = _ref.model,
        reducerEnhancer = _ref.reducerEnhancer,
        references = _ref.references;

    function simpleProduce(path, state, fn) {
      if (disableImmer) {
        var _current = get(path, state);

        var next = fn(_current);

        if (_current !== next) {
          return newify(path, state, next);
        }

        return state;
      }

      var draft = immerPeasy.createDraft(state);

      var current = get(path, draft);

      fn(current);
      return immerPeasy.finishDraft(draft);
    }

    var defaultState = initialState;
    var actionCreatorDict = {};
    var actionCreators = {};
    var actionReducersDict = {};
    var actionThunks = {};
    var computedProperties = [];
    var customReducers = [];
    var listenerActionCreators = {};
    var listenerActionMap = {};
    var listenerDefinitions = [];
    var computedState = {
      isInReducer: false,
      currentState: defaultState
    };

    var recursiveExtractDefsFromModel = function recursiveExtractDefsFromModel(current, parentPath) {
      return Object.keys(current).forEach(function (key) {
        var value = current[key];
        var path = [].concat(parentPath, [key]);
        var meta = {
          parent: parentPath,
          path: path
        };

        var handleValueAsState = function handleValueAsState() {
          var initialParentRef = get(parentPath, initialState);

          if (initialParentRef && key in initialParentRef) {
            set(path, defaultState, initialParentRef[key]);
          } else {
            set(path, defaultState, value);
          }
        };

        if (typeof value === 'function') {
          if (value[actionSymbol] || value[actionOnSymbol]) {
            var prefix = value[actionSymbol] ? '@action' : '@actionOn';
            var type = prefix + "." + path.join('.');
            var actionMeta = value[actionSymbol] || value[actionOnSymbol];
            actionMeta.actionName = key;
            actionMeta.type = type;
            actionMeta.parent = meta.parent;
            actionMeta.path = meta.path; // Action Reducer

            actionReducersDict[type] = value; // Action Creator

            var actionCreator = function actionCreator(payload) {
              var actionDefinition = {
                type: type,
                payload: payload
              };

              if (value[actionOnSymbol] && actionMeta.resolvedTargets) {
                payload.resolvedTargets = [].concat(actionMeta.resolvedTargets);
              }

              var result = references.dispatch(actionDefinition);
              return result;
            };

            actionCreator.type = type;
            actionCreatorDict[type] = actionCreator;

            if (key !== 'easyPeasyReplaceState') {
              if (value[actionOnSymbol]) {
                listenerDefinitions.push(value);
                set(path, listenerActionCreators, actionCreator);
              } else {
                set(path, actionCreators, actionCreator);
              }
            }
          } else if (value[thunkSymbol] || value[thunkOnSymbol]) {
            var _prefix = value[thunkSymbol] ? '@thunk' : '@thunkOn';

            var _type = _prefix + "." + path.join('.');

            var thunkMeta = value[thunkSymbol] || value[thunkOnSymbol];
            thunkMeta.actionName = key;
            thunkMeta.type = _type;
            thunkMeta.parent = meta.parent;
            thunkMeta.path = meta.path; // Thunk Action

            var thunkHandler = function thunkHandler(payload) {
              var helpers = {
                dispatch: references.dispatch,
                getState: function getState() {
                  return get(parentPath, references.getState());
                },
                getStoreActions: function getStoreActions() {
                  return actionCreators;
                },
                getStoreState: references.getState,
                injections: injections,
                meta: meta
              };

              if (value[thunkOnSymbol] && thunkMeta.resolvedTargets) {
                payload.resolvedTargets = [].concat(thunkMeta.resolvedTargets);
              }

              return value(get(parentPath, actionCreators), payload, helpers);
            };

            set(path, actionThunks, thunkHandler); // Thunk Action Creator

            var startType = _type + "(start)";
            var successType = _type + "(success)";
            var failType = _type + "(fail)";

            var _actionCreator = function _actionCreator(payload) {
              var dispatchError = function dispatchError(err) {
                references.dispatch({
                  type: failType,
                  payload: payload,
                  error: err
                });
                references.dispatch({
                  type: _type,
                  payload: payload,
                  error: err
                });
              };

              var dispatchSuccess = function dispatchSuccess(result) {
                references.dispatch({
                  type: successType,
                  payload: payload,
                  result: result
                });
                references.dispatch({
                  type: _type,
                  payload: payload,
                  result: result
                });
              };

              references.dispatch({
                type: startType,
                payload: payload
              });

              try {
                var result = references.dispatch(function () {
                  return thunkHandler(payload);
                });

                if (typeof result === 'object' && typeof result.then === 'function') {
                  return result.then(function (resolved) {
                    dispatchSuccess(resolved);
                    return resolved;
                  }).catch(function (err) {
                    dispatchError(err);
                    throw err;
                  });
                }

                dispatchSuccess(result);
                return result;
              } catch (err) {
                dispatchError(err);
                throw err;
              }
            };

            _actionCreator.type = _type;
            _actionCreator.startType = startType;
            _actionCreator.successType = successType;
            _actionCreator.failType = failType;
            actionCreatorDict[_type] = _actionCreator;

            if (value[thunkOnSymbol]) {
              listenerDefinitions.push(value);
              set(path, listenerActionCreators, _actionCreator);
            } else {
              set(path, actionCreators, _actionCreator);
            }
          } else if (value[computedSymbol]) {
            var parent = get(parentPath, defaultState);

            var computedMeta = value[computedSymbol];
            var memoisedResultFn = memoizerific(1)(value);

            var createComputedProperty = function createComputedProperty(o) {
              Object.defineProperty(o, key, {
                configurable: true,
                enumerable: true,
                get: function get$1() {
                  var storeState;

                  if (computedState.isInReducer) {
                    storeState = computedState.currentState;
                  } else if (references.getState == null) {
                    return undefined;
                  } else {
                    storeState = references.getState();
                  }

                  var state = get(parentPath, storeState);

                  var inputs = computedMeta.stateResolvers.map(function (resolver) {
                    return resolver(state, storeState);
                  });
                  return memoisedResultFn.apply(void 0, inputs);
                }
              });
            };

            createComputedProperty(parent);
            computedProperties.push({
              key: key,
              parentPath: parentPath,
              createComputedProperty: createComputedProperty
            });
          } else if (value[reducerSymbol]) {
            customReducers.push({
              key: key,
              parentPath: parentPath,
              reducer: value
            });
          } else {
            handleValueAsState();
          }
        } else if (isStateObject(value)) {
          var existing = get(path, defaultState);

          if (existing == null) {
            set(path, defaultState, {});
          }

          recursiveExtractDefsFromModel(value, path);
        } else {
          handleValueAsState();
        }
      });
    };

    recursiveExtractDefsFromModel(model, []);
    listenerDefinitions.forEach(function (listenerActionOrThunk) {
      var listenerMeta = listenerActionOrThunk[actionOnSymbol] || listenerActionOrThunk[thunkOnSymbol];
      var targets = listenerMeta.targetResolver(get(listenerMeta.parent, actionCreators), actionCreators);
      var targetTypes = (Array.isArray(targets) ? targets : [targets]).reduce(function (acc, target) {
        if (typeof target === 'function' && target.type && actionCreatorDict[target.type]) {
          acc.push(target.type);
        } else if (typeof target === 'string') {
          acc.push(target);
        }

        return acc;
      }, []);
      listenerMeta.resolvedTargets = targetTypes;
      targetTypes.forEach(function (targetType) {
        var listenerReg = listenerActionMap[targetType] || [];
        listenerReg.push(actionCreatorDict[listenerMeta.type]);
        listenerActionMap[targetType] = listenerReg;
      });
    });

    var createReducer = function createReducer() {
      var runActionReducerAtPath = function runActionReducerAtPath(state, action, actionReducer, path) {
        return simpleProduce(path, state, function (draft) {
          return actionReducer(draft, action.payload);
        });
      };

      var reducerForActions = function reducerForActions(state, action) {
        var actionReducer = actionReducersDict[action.type];

        if (actionReducer) {
          var actionMeta = actionReducer[actionSymbol] || actionReducer[actionOnSymbol];
          return runActionReducerAtPath(state, action, actionReducer, actionMeta.parent);
        }

        return state;
      };

      var reducerForCustomReducers = function reducerForCustomReducers(state, action) {
        return customReducers.reduce(function (acc, _ref2) {
          var parentPath = _ref2.parentPath,
              key = _ref2.key,
              red = _ref2.reducer;
          return simpleProduce(parentPath, acc, function (draft) {
            draft[key] = red(draft[key], action);
            return draft;
          });
        }, state);
      };

      var rootReducer = function rootReducer(state, action) {
        var stateAfterActions = reducerForActions(state, action);
        var next = customReducers.length > 0 ? reducerForCustomReducers(stateAfterActions, action) : stateAfterActions;

        if (state !== next) {
          computedProperties.forEach(function (_ref3) {
            var parentPath = _ref3.parentPath,
                createComputedProperty = _ref3.createComputedProperty;
            createComputedProperty(get(parentPath, next));
          });
        }

        return next;
      };

      return rootReducer;
    };

    return {
      actionCreatorDict: actionCreatorDict,
      actionCreators: actionCreators,
      computedProperties: computedProperties,
      computedState: computedState,
      defaultState: defaultState,
      listenerActionCreators: listenerActionCreators,
      listenerActionMap: listenerActionMap,
      reducer: reducerEnhancer(createReducer())
    };
  }

  function createStore(model, options) {
    if (options === void 0) {
      options = {};
    }

    var _options = options,
        compose = _options.compose,
        _options$devTools = _options.devTools,
        devTools = _options$devTools === void 0 ? true : _options$devTools,
        _options$disableImmer = _options.disableImmer,
        disableImmer = _options$disableImmer === void 0 ? false : _options$disableImmer,
        _options$enhancers = _options.enhancers,
        enhancers = _options$enhancers === void 0 ? [] : _options$enhancers,
        _options$initialState = _options.initialState,
        initialState = _options$initialState === void 0 ? {} : _options$initialState,
        injections = _options.injections,
        _options$middleware = _options.middleware,
        middleware = _options$middleware === void 0 ? [] : _options$middleware,
        _options$mockActions = _options.mockActions,
        mockActions = _options$mockActions === void 0 ? false : _options$mockActions,
        _options$name = _options.name,
        storeName = _options$name === void 0 ? "EasyPeasyStore" : _options$name,
        _options$reducerEnhan = _options.reducerEnhancer,
        reducerEnhancer = _options$reducerEnhan === void 0 ? function (rootReducer) {
      return rootReducer;
    } : _options$reducerEnhan;

    var bindReplaceState = function bindReplaceState(modelDef) {
      return _extends({}, modelDef, {
        easyPeasyReplaceState: action(function (state, payload) {
          return payload;
        })
      });
    };

    var modelDefinition = bindReplaceState(model);
    var mockedActions = [];
    var references = {};

    var bindStoreInternals = function bindStoreInternals(state) {
      if (state === void 0) {
        state = {};
      }

      references.internals = createStoreInternals({
        disableImmer: disableImmer,
        initialState: state,
        injections: injections,
        model: modelDefinition,
        reducerEnhancer: reducerEnhancer,
        references: references
      });
    };

    bindStoreInternals(initialState);

    var listenerActionsMiddleware = function listenerActionsMiddleware() {
      return function (next) {
        return function (action) {
          var result = next(action);

          if (action && references.internals.listenerActionMap[action.type] && references.internals.listenerActionMap[action.type].length > 0) {
            var sourceAction = references.internals.actionCreatorDict[action.type];
            references.internals.listenerActionMap[action.type].forEach(function (actionCreator) {
              actionCreator({
                type: sourceAction ? sourceAction.type : action.type,
                payload: action.payload,
                error: action.error,
                result: action.result
              });
            });
          }

          return result;
        };
      };
    };

    var mockActionsMiddleware = function mockActionsMiddleware() {
      return function () {
        return function (action) {
          if (action != null) {
            mockedActions.push(action);
          }

          return undefined;
        };
      };
    };

    var computedPropertiesMiddleware = function computedPropertiesMiddleware(store) {
      return function (next) {
        return function (action) {
          references.internals.computedState.currentState = store.getState();
          references.internals.computedState.isInReducer = true;
          return next(action);
        };
      };
    };

    var easyPeasyMiddleware = [computedPropertiesMiddleware, reduxThunk].concat(middleware, [listenerActionsMiddleware]);

    if (mockActions) {
      easyPeasyMiddleware.push(mockActionsMiddleware);
    }

    var composeEnhancers = compose || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__({
      name: storeName
    }) : redux.compose);
    var store = redux.createStore(references.internals.reducer, references.internals.defaultState, composeEnhancers.apply(void 0, [redux.applyMiddleware.apply(void 0, easyPeasyMiddleware)].concat(enhancers)));
    store.subscribe(function () {
      references.internals.computedState.isInReducer = false;
    });
    references.dispatch = store.dispatch;
    references.getState = store.getState;

    var bindActionCreators = function bindActionCreators() {
      Object.keys(store.dispatch).forEach(function (actionsKey) {
        delete store.dispatch[actionsKey];
      });
      Object.keys(references.internals.actionCreators).forEach(function (key) {
        store.dispatch[key] = references.internals.actionCreators[key];
      });
    };

    bindActionCreators();

    var rebindStore = function rebindStore(removeKey) {
      var currentState = store.getState();

      if (removeKey) {
        delete currentState[removeKey];
      }

      bindStoreInternals(store.getState());
      store.replaceReducer(references.internals.reducer);
      references.internals.actionCreatorDict['@action.easyPeasyReplaceState'](references.internals.defaultState);
      bindActionCreators();
    };

    return Object.assign(store, {
      addModel: function addModel(key, modelForKey) {
        if (modelDefinition[key] && "development" !== 'production') {
          // eslint-disable-next-line no-console
          console.warn("easy-peasy: The store model already contains a model definition for \"" + key + "\"");
          store.removeModel(key);
        }

        modelDefinition[key] = modelForKey;
        rebindStore();
      },
      clearMockedActions: function clearMockedActions() {
        mockedActions = [];
      },
      getActions: function getActions() {
        return references.internals.actionCreators;
      },
      getListeners: function getListeners() {
        return references.internals.listenerActionCreators;
      },
      getMockedActions: function getMockedActions() {
        return [].concat(mockedActions);
      },
      reconfigure: function reconfigure(newModel) {
        modelDefinition = bindReplaceState(newModel);
        rebindStore();
      },
      removeModel: function removeModel(key) {
        if (!modelDefinition[key]) {
          {
            // eslint-disable-next-line no-console
            console.warn("easy-peasy: The store model does not contain a model definition for \"" + key + "\"");
          }

          return;
        }

        delete modelDefinition[key];
        rebindStore(key);
      }
    });
  }

  /* eslint-disable react/prop-types */
  function createContextStore(model, config) {
    var StoreContext = React.createContext();

    function Provider(_ref) {
      var children = _ref.children,
          initialData = _ref.initialData;
      var store = React.useMemo(function () {
        return createStore(typeof model === 'function' ? model(initialData) : model, config);
      }, []);
      return React__default.createElement(StoreContext.Provider, {
        value: store
      }, children);
    }

    function useStore() {
      return React.useContext(StoreContext);
    }

    return {
      Provider: Provider,
      useStore: useStore,
      useStoreState: createStoreStateHook(StoreContext),
      useStoreActions: createStoreActionsHook(StoreContext),
      useStoreDispatch: createStoreDispatchHook(StoreContext)
    };
  }

  /**
   * Some good references on the topic of reinitialisation:
   * - https://github.com/facebook/react/issues/14830
   */

  function createComponentStore(model, config) {
    return function useLocalStore(initialData) {
      var store = React.useMemo(function () {
        return createStore(typeof model === 'function' ? model(initialData) : model, config);
      }, []);
      var previousStateRef = React.useRef(store.getState());

      var _useState = React.useState(function () {
        return store.getState();
      }),
          currentState = _useState[0],
          setCurrentState = _useState[1];

      React.useEffect(function () {
        return store.subscribe(function () {
          var nextState = store.getState();

          if (previousStateRef.current !== nextState) {
            previousStateRef.current = nextState;
            setCurrentState(nextState);
          }
        });
      }, [store]);
      return [currentState, store.getActions()];
    };
  }

  var StoreProvider = function StoreProvider(_ref) {
    var children = _ref.children,
        store = _ref.store;
    return React__default.createElement(StoreContext.Provider, {
      value: store
    }, children);
  };

  /**
   * The auto freeze feature of immer doesn't seem to work in our testing. We have
   * explicitly disabled it to avoid perf issues.
   */

  immerPeasy.setAutoFreeze(false);

  exports.StoreProvider = StoreProvider;
  exports.action = action;
  exports.actionOn = actionOn;
  exports.computed = computed;
  exports.createComponentStore = createComponentStore;
  exports.createContextStore = createContextStore;
  exports.createStore = createStore;
  exports.createTypedHooks = createTypedHooks;
  exports.debug = debug;
  exports.memo = memo;
  exports.reducer = reducer;
  exports.thunk = thunk;
  exports.thunkOn = thunkOn;
  exports.useStoreActions = useStoreActions;
  exports.useStoreDispatch = useStoreDispatch;
  exports.useStoreState = useStoreState;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=easy-peasy.umd.development.js.map
